from __future__ import print_function as _
from pysb.importers.bngl import model_from_bngl
import pysb.pathfinder as pf
import subprocess
import os
import tempfile
import shutil
import re
try:
    # Python 3
    from urllib.request import urlretrieve
except ImportError:
    # Python 2
    from urllib import urlretrieve
from pysb.logging import get_logger, EXTENDED_DEBUG

BIOMODELS_REGEX = re.compile(r'(BIOMD|MODEL)[0-9]{10}')
BIOMODELS_URLS = {
    'ebi': 'http://www.ebi.ac.uk/biomodels-main/download?mid={}',
    'caltech': 'http://biomodels.caltech.edu/download?mid={}'
}


class SbmlTranslationError(Exception):
    pass


def sbml_translator(input_file,
                    output_file=None,
                    convention_file=None,
                    naming_conventions=None,
                    user_structures=None,
                    molecule_id=False,
                    atomize=False,
                    pathway_commons=False,
                    verbose=False):
    """
    Run the BioNetGen sbmlTranslator binary to convert SBML to BNGL

    This function runs the external program sbmlTranslator, included with
    BioNetGen, which converts SBML files to BioNetGen language (BNGL). If
    PySB was installed using "conda", you can install sbmlTranslator using
    "conda install -c alubbock atomizer". sbmlTranslator is bundled with
    BioNetGen if BNG is installed by manual download and unzip.

    Generally, PySB users don't need to run this function directly; an SBML
    model can be imported to PySB in a single step with
    :func:`model_from_sbml`. However, users may wish to note the parameters
    for this function, which alter the way the SBML file is processed. These
    parameters can be supplied as ``**kwargs`` to :func:`model_from_sbml`.

    For more detailed descriptions of the arguments, see the `sbmlTranslator
    documentation <http://bionetgen.org/index.php/SBML2BNGL>`_.

    Parameters
    ----------
    input_file : string
        SBML input filename
    output_file : string, optional
        BNGL output filename
    convention_file : string, optional
        Conventions filename
    naming_conventions : string, optional
        Naming conventions filename
    user_structures : string, optional
        User structures filename
    molecule_id : bool, optional
        Use SBML molecule IDs (True) or names (False).
        IDs are less descriptive but more BNGL friendly. Use only if the
        generated BNGL has syntactic errors
    atomize : bool, optional
        Atomize the model, i.e. attempt to infer molecular structure and
        build rules from the model (True) or just perform a flat import (False)
    pathway_commons : bool, optional
        Use pathway commons to infer molecule binding. This
        setting requires an internet connection and will query the pathway
        commons web service.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False leaves the logging verbosity unchanged, True is equal
        to DEBUG.

    Returns
    -------
    string
        BNGL output filename
    """
    logger = get_logger(__name__, log_level=verbose)
    sbmltrans_bin = pf.get_path('atomizer')

    sbmltrans_args = [sbmltrans_bin, '-i', input_file]
    if output_file is None:
        output_file = os.path.splitext(input_file)[0] + '.bngl'
    sbmltrans_args.extend(['-o', output_file])

    if convention_file:
        sbmltrans_args.extend(['-c', convention_file])

    if naming_conventions:
        sbmltrans_args.extend(['-n', naming_conventions])

    if user_structures:
        sbmltrans_args.extend(['-u', user_structures])

    if molecule_id:
        sbmltrans_args.append('-id')

    if atomize:
        sbmltrans_args.append('-a')

    if pathway_commons:
        sbmltrans_args.append('-p')

    logger.debug("sbmlTranslator command: " + " ".join(sbmltrans_args))

    p = subprocess.Popen(sbmltrans_args,
                         cwd=os.getcwd(),
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)

    if logger.getEffectiveLevel() <= EXTENDED_DEBUG:
        output = "\n".join([line for line in iter(p.stdout.readline, b'')])
        if output:
            logger.log(EXTENDED_DEBUG, "sbmlTranslator output:\n\n" + output)
    (p_out, p_err) = p.communicate()
    if p.returncode:
        raise SbmlTranslationError(p_out.decode('utf-8') + "\n" +
                                   p_err.decode('utf-8'))

    return output_file


def model_from_sbml(filename, force=False, cleanup=True, **kwargs):
    """
    Create a PySB Model object from an Systems Biology Markup Language (SBML)
    file, using BioNetGen's
    `sbmlTranslator <http://bionetgen.org/index.php/SBML2BNGL>`_,
    which can attempt to extrapolate higher-level (rule-based) structure
    from an SBML source file (argument atomize=True). The model is first
    converted into BioNetGen language by sbmlTranslator, then PySB's
    :class:`BnglBuilder` class converts the BioNetGen language model into a
    PySB Model.

    Notes
    -----

    Requires the sbmlTranslator program (also known at Atomizer). If
    PySB was installed using "conda", you can install sbmlTranslator using
    "conda install -c alubbock atomizer". It is bundled with BioNetGen if
    BNG is installed by manual download and unzip.

    Read the `sbmlTranslator documentation
    <http://bionetgen.org/index.php/SBML2BNGL>`_ for further information on
    sbmlTranslator's limitations.

    Parameters
    ----------
    filename :
        A Systems Biology Markup Language .sbml file
    force : bool, optional
        The default, False, will raise an Exception if there are any errors
        importing the model to PySB, e.g. due to unsupported features.
        Setting to True will attempt to ignore any import errors, which may
        lead to a model that only poorly represents the original. Use at own
        risk!
    cleanup : bool
        Delete temporary directory on completion if True. Set to False for
        debugging purposes.
    **kwargs: kwargs
        Keyword arguments to pass on to :func:`sbml_translator`
    """
    logger = get_logger(__name__, log_level=kwargs.get('verbose'))
    tmpdir = tempfile.mkdtemp()
    logger.debug("Performing SBML to BNGL translation in temporary "
                 "directory %s" % tmpdir)
    try:
        bngl_file = os.path.join(tmpdir, 'model.bngl')
        sbml_translator(filename, bngl_file, **kwargs)
        return model_from_bngl(bngl_file, force=force, cleanup=cleanup)
    finally:
        if cleanup:
            shutil.rmtree(tmpdir)


def model_from_biomodels(accession_no, force=False, cleanup=True,
                         mirror='ebi', **kwargs):
    """
    Create a PySB Model based on a BioModels SBML model

    Downloads file from BioModels (https://www.ebi.ac.uk/biomodels-main/)
    and runs it through :func:`model_from_sbml`. See that function for
    further details on additional arguments and implementation details.
    Utilizes BioNetGen's SBMLTranslator.

    Notes
    -----

    Requires the sbmlTranslator program (also known at Atomizer). If
    PySB was installed using "conda", you can install sbmlTranslator using
    "conda install -c alubbock atomizer". It is bundled with BioNetGen if
    BNG is installed by manual download and unzip.

    Read the `sbmlTranslator documentation
    <http://bionetgen.org/index.php/SBML2BNGL>`_ for further information on
    sbmlTranslator's limitations.

    Parameters
    ----------
    accession_no : str
        A BioModels accession number - the string 'BIOMD' followed by 10
        digits, e.g. 'BIOMD0000000001'. For brevity, just the last digits will
        be accepted as a string, e.g. '1' is equivalent the accession number
        in the previous sentence.
    force : bool, optional
        The default, False, will raise an Exception if there are any errors
        importing the model to PySB, e.g. due to unsupported features.
        Setting to True will attempt to ignore any import errors, which may
        lead to a model that only poorly represents the original. Use at own
        risk!
    cleanup : bool
        Delete temporary directory on completion if True. Set to False for
        debugging purposes.
    mirror : str
        Which BioModels mirror to use, either 'ebi' or 'caltech'
    **kwargs: kwargs
        Keyword arguments to pass on to :func:`sbml_translator`

    Examples
    --------

    >>> from pysb.importers.sbml import model_from_biomodels
    >>> model = model_from_biomodels('1')           #doctest: +SKIP
    >>> print(model)                                #doctest: +SKIP
    <Model 'pysb' (monomers: 12, rules: 17, parameters: 37, expressions: 0, ...
    """
    logger = get_logger(__name__, log_level=kwargs.get('verbose'))
    if not BIOMODELS_REGEX.match(accession_no):
        try:
            accession_no = 'BIOMD{:010d}'.format(int(accession_no))
        except ValueError:
            raise ValueError('accession_no must be an integer or a BioModels '
                             'accession number (BIOMDxxxxxxxxxx)')
    logger.info('Importing model {} to PySB'.format(accession_no))
    filename = _download_biomodels(accession_no, mirror=mirror)
    try:
        return model_from_sbml(filename, force=force, cleanup=cleanup,
                               **kwargs)
    finally:
        try:
            os.remove(filename)
        except OSError:
            pass


def _download_biomodels(accession_no, mirror):
    try:
        url_fmt = BIOMODELS_URLS[mirror]
    except KeyError:
        raise ValueError('Unknown Biomodels mirror: "{}". Choices are: {}'
                         .format(mirror, BIOMODELS_URLS.keys()))
    filename, _ = urlretrieve(url_fmt.format(accession_no))
    return filename



"""
Module containing a class for exporting a PySB model to SBML using libSBML

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""
import pysb
import pysb.bng
from pysb.export import Exporter
from sympy.printing.mathml import MathMLPrinter
from sympy import Symbol
from xml.dom.minidom import Document
import itertools
try:
    import libsbml
except ImportError:
    libsbml = None


class MathMLContentPrinter(MathMLPrinter):
    """Prints an expression to MathML without presentation markup."""
    def _print_Symbol(self, sym):
        ci = self.dom.createElement(self.mathml_tag(sym))
        ci.appendChild(self.dom.createTextNode(sym.name))
        return ci

    def to_xml(self, expr):
        # Preferably this should use a public API, but as that doesn't exist...
        return self._print(expr)


def _check(value):
    """
    Validate a libsbml return value

    Raises ValueError if 'value' is a libsbml error code or None.
    """
    if type(value) is int and value != libsbml.LIBSBML_OPERATION_SUCCESS:
        raise ValueError(
            'Error encountered converting to SBML. '
            'LibSBML returned error code {}: "{}"'.format(
                value,
                libsbml.OperationReturnValue_toString(value).strip()
            )
        )
    elif value is None:
        raise ValueError('LibSBML returned a null value')


def _add_ci(x_doc, x_parent, name):
    """ Add <ci>name</ci> element to <x_parent> within x_doc """
    ci = x_doc.createElement('ci')
    ci.appendChild(x_doc.createTextNode(name))
    x_parent.appendChild(ci)


def _xml_to_ast(x_element):
    """ Wrap MathML fragment with <math> tag and convert to libSBML AST """
    x_doc = Document()
    x_mathml = x_doc.createElement('math')
    x_mathml.setAttribute('xmlns', 'http://www.w3.org/1998/Math/MathML')

    x_mathml.appendChild(x_element)
    x_doc.appendChild(x_mathml)

    mathml_ast = libsbml.readMathMLFromString(x_doc.toxml())
    _check(mathml_ast)
    return mathml_ast


def _mathml_expr_call(expr):
    """ Generate an XML <apply> expression call """
    x_doc = Document()
    x_apply = x_doc.createElement('apply')
    x_doc.appendChild(x_apply)

    _add_ci(x_doc, x_apply, expr.name)
    for sym in expr.expand_expr(expand_observables=True).free_symbols:
        if isinstance(sym, pysb.Expression):
            continue
        _add_ci(x_doc, x_apply, sym.name if isinstance(sym, pysb.Parameter) else str(sym))

    return x_apply


class SbmlExporter(Exporter):
    """A class for returning the SBML for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """
    def __init__(self, *args, **kwargs):
        if not libsbml:
            raise ImportError('The SbmlExporter requires the libsbml python package')
        super(SbmlExporter, self).__init__(*args, **kwargs)

    def _sympy_to_sbmlast(self, sympy_expr):
        """
        Convert a sympy expression to the AST format used by libsbml
        """
        return _xml_to_ast(MathMLContentPrinter().to_xml(sympy_expr))

    def convert(self, level=(3, 2)):
        """
        Convert the PySB model to a libSBML document

        Requires the libsbml python package

        Parameters
        ----------
        level: (int, int)
            The SBML level and version to use. The default is SBML level 3, version 2. Conversion
            to other levels/versions may not be possible or may lose fidelity.

        Returns
        -------
        libsbml.SBMLDocument
            A libSBML document converted form the PySB model
        """
        doc = libsbml.SBMLDocument(3, 2)
        smodel = doc.createModel()
        _check(smodel)
        _check(smodel.setName(self.model.name))

        pysb.bng.generate_equations(self.model)

        # Docstring
        if self.docstring:
            notes_str = """
            <notes>
                <body xmlns="http://www.w3.org/1999/xhtml">
                    <p>%s</p>
                </body>
            </notes>""" % self.docstring.replace("\n", "<br />\n"+" "*20)
            _check(smodel.setNotes(notes_str))

        # Compartments
        if self.model.compartments:
            for cpt in self.model.compartments:
                c = smodel.createCompartment()
                _check(c)
                _check(c.setId(cpt.name))
                _check(c.setSpatialDimensions(cpt.dimension))
                _check(c.setSize(cpt.size.value))
                _check(c.setConstant(True))
        else:
            c = smodel.createCompartment()
            _check(c)
            _check(c.setId('default'))
            _check(c.setSpatialDimensions(3))
            _check(c.setSize(1))
            _check(c.setConstant(True))

        # Expressions
        for expr in itertools.chain(
                self.model.expressions_constant(),
                self.model.expressions_dynamic(include_local=False),
                self.model._derived_expressions
        ):
            # create an observable "parameter"
            e = smodel.createParameter()
            _check(e)
            _check(e.setId(expr.name))
            _check(e.setName(expr.name))
            _check(e.setConstant(False))

            # create an assignment rule which assigns the expression to the parameter
            expr_rule = smodel.createAssignmentRule()

            _check(expr_rule)
            _check(expr_rule.setVariable(e.getId()))

            expr_mathml = self._sympy_to_sbmlast(expr.expand_expr(expand_observables=True))
            _check(expr_rule.setMath(expr_mathml))

        # Initial values/assignments
        fixed_species_idx = set()
        initial_species_idx = set()
        for ic in self.model.initials:
            sp_idx = self.model.get_species_index(ic.pattern)
            ia = smodel.createInitialAssignment()
            _check(ia)
            _check(ia.setSymbol('__s{}'.format(sp_idx)))
            init_mathml = self._sympy_to_sbmlast(Symbol(ic.value.name))
            _check(ia.setMath(init_mathml))
            initial_species_idx.add(sp_idx)

            if ic.fixed:
                fixed_species_idx.add(sp_idx)

        # Species
        for i, s in enumerate(self.model.species):
            sp = smodel.createSpecies()
            _check(sp)
            _check(sp.setId('__s{}'.format(i)))
            if self.model.compartments:
                # Try to determine compartment, which must be unique for the species
                mon_cpt = set(mp.compartment for mp in s.monomer_patterns if mp.compartment is not None)
                if len(mon_cpt) == 0 and s.compartment:
                    compartment_name = s.compartment_name
                elif len(mon_cpt) == 1:
                    mon_cpt = mon_cpt.pop()
                    if s.compartment is not None and mon_cpt != s.compartment:
                        raise ValueError('Species {} has different monomer and species compartments, '
                                         'which is not supported in SBML'.format(s))
                    compartment_name = mon_cpt.name
                else:
                    raise ValueError('Species {} has more than one different monomer compartment, '
                                     'which is not supported in SBML'.format(s))
            else:
                compartment_name = 'default'
            _check(sp.setCompartment(compartment_name))
            _check(sp.setName(str(s).replace('% ', '._br_')))
            _check(sp.setBoundaryCondition(i in fixed_species_idx))
            _check(sp.setConstant(False))
            _check(sp.setHasOnlySubstanceUnits(True))
            if i not in initial_species_idx:
                _check(sp.setInitialAmount(0.0))


        # Parameters

        for param in itertools.chain(self.model.parameters,
                                     self.model._derived_parameters):
            p = smodel.createParameter()
            _check(p)
            _check(p.setId(param.name))
            _check(p.setName(param.name))
            _check(p.setValue(param.value))
            _check(p.setConstant(True))

        # Reactions
        for i, reaction in enumerate(self.model.reactions_bidirectional):
            rxn = smodel.createReaction()
            _check(rxn)
            _check(rxn.setId('r{}'.format(i)))
            _check(rxn.setName(' + '.join(reaction['rule'])))
            _check(rxn.setReversible(reaction['reversible']))

            for sp in reaction['reactants']:
                reac = rxn.createReactant()
                _check(reac)
                _check(reac.setSpecies('__s{}'.format(sp)))
                _check(reac.setConstant(True))

            for sp in reaction['products']:
                prd = rxn.createProduct()
                _check(prd)
                _check(prd.setSpecies('__s{}'.format(sp)))
                _check(prd.setConstant(True))

            for symbol in reaction['rate'].free_symbols:
                if isinstance(symbol, pysb.Expression):
                    expr = symbol.expand_expr(expand_observables=True)
                    for sym in expr.free_symbols:
                        if not isinstance(sym, (pysb.Parameter, pysb.Expression)):
                            # Species reference, needs to be specified as modifier
                            modifier = rxn.createModifier()
                            _check(modifier)
                            _check(modifier.setSpecies(str(sym)))

            rate = rxn.createKineticLaw()
            _check(rate)
            rate_mathml = self._sympy_to_sbmlast(reaction['rate'])
            _check(rate.setMath(rate_mathml))

        # Observables
        for i, observable in enumerate(self.model.observables):
            # create an observable "parameter"
            obs = smodel.createParameter()
            _check(obs)
            _check(obs.setId('__obs{}'.format(i)))
            _check(obs.setName(observable.name))
            _check(obs.setConstant(False))

            # create an assignment rule which assigns the observable expression to the parameter
            obs_rule = smodel.createAssignmentRule()

            _check(obs_rule)
            _check(obs_rule.setVariable(obs.getId()))

            obs_mathml = self._sympy_to_sbmlast(observable.expand_obs())
            _check(obs_rule.setMath(obs_mathml))




        # Apply any requested level/version conversion
        if level != (3, 2):
            prop = libsbml.ConversionProperties(libsbml.SBMLNamespaces(*level))
            prop.addOption('strict', False)
            prop.addOption('setLevelAndVersion', True)
            prop.addOption('ignorePackages', True)
            _check(doc.convert(prop))

        return doc

    def export(self, level=(3, 2)):
        """
        Export the SBML for the PySB model associated with the exporter

        Requires libsbml package.

        Parameters
        ----------
        level: (int, int)
            The SBML level and version to use. The default is SBML level 3, version 2. Conversion
            to other levels/versions may not be possible or may lose fidelity.

        Returns
        -------
        string
            String containing the SBML output.
        """
        return libsbml.writeSBMLToString(self.convert(level=level))



"""Translation of the BioNetGen example "egfr_simple" from the BNG wiki.

http://bionetgen.org/index.php/Egfr_simple
"""

from __future__ import print_function
from pysb import *

Model()


Parameter('NA', 6.02e23)     # Avogadro's number (molecules/mol)
Parameter('f', 1)            # Fraction of the cell to simulate
Expression('Vo', f*1.0e-10)  # Extracellular volume=1/cell_density (L)
Expression('V', f*3.0e-12)   # Cytoplasmic volume (L)

# Initial amount of ligand (20 nM) converted to copies per cell
Expression('EGF_init', 20*1e-9*NA*Vo)

# Initial amounts of cellular components (copies per cell)
Expression('EGFR_init', f*1.8e5)
Expression('Grb2_init', f*1.5e5)
Expression('Sos1_init', f*6.2e4)

# Rate constants
# Divide by NA*V to convert bimolecular rate constants from /M/sec to
# /(molecule/cell)/sec
Expression('kp1', 9.0e7/(NA*Vo))  # ligand-monomer binding
Parameter('km1', 0.06)            # ligand-monomer dissociation
Expression('kp2', 1.0e7/(NA*V))   # aggregation of bound monomers
Parameter('km2', 0.1)             # dissociation of bound monomers
Parameter('kp3', 0.5)             # dimer transphosphorylation
Parameter('km3', 4.505)           # dimer dephosphorylation
Expression('kp4', 1.5e6/(NA*V))   # binding of Grb2 to receptor
Parameter('km4', 0.05)            # dissociation of Grb2 from receptor
Expression('kp5', 1.0e7/(NA*V))   # binding of Grb2 to Sos1
Parameter('km5', 0.06)            # dissociation of Grb2 from Sos1
Parameter('deg', 0.01)            # degradation of receptor dimers


Monomer('EGF', ['R'])
Monomer('EGFR', ['L','CR1','Y1068'], {'Y1068':('U','P')})
Monomer('Grb2', ['SH2','SH3'])
Monomer('Sos1', ['PxxP'])


Initial(EGF(R=None), EGF_init)
Initial(EGFR(L=None, CR1=None, Y1068='U'), EGFR_init)
Initial(Grb2(SH2=None, SH3=None), Grb2_init)
Initial(Sos1(PxxP=None), Sos1_init)


Observable('EGFR_tot', EGFR())
Observable('Lig_free', EGF(R=None))
Observable('Dim', EGFR(CR1=ANY), match='species')
Observable('RP', EGFR(Y1068=('P',WILD)))
Observable('Grb2Sos1', Grb2(SH2=None, SH3=1) % Sos1(PxxP=1))
Observable('Sos1_act', EGFR(Y1068=1) % Grb2(SH2=1, SH3=2) % Sos1(PxxP=2))


# Ligand-receptor binding
Rule('egf_bind_egfr',
     EGFR(L=None, CR1=None) + EGF(R=None) | EGFR(L=1, CR1=None) % EGF(R=1),
     kp1, km1)

# Receptor-aggregation
Rule('egfr_dimerize',
     EGFR(L=ANY, CR1=None) + EGFR(L=ANY, CR1=None) |
     EGFR(L=ANY, CR1=1) % EGFR(L=ANY, CR1=1),
     kp2, km2)

# Transphosphorylation of EGFR by RTK
Rule('egfr_transphos',
     EGFR(CR1=ANY, Y1068='U') >> EGFR(CR1=ANY, Y1068='P'), kp3)

# Dephosphorylation
Rule('egfr_dephos',
     EGFR(Y1068='P') >> EGFR(Y1068='U'), km3)

# Grb2 binding to pY1068
Rule('grb2_bind_egfr',
     EGFR(Y1068='P') + Grb2(SH2=None) | EGFR(Y1068=('P',1)) % Grb2(SH2=1),
     kp4, km4)

# Grb2 binding to Sos1
Rule('sos1_bind_grb2',
     Grb2(SH3=None) + Sos1(PxxP=None) | Grb2(SH3=1) % Sos1(PxxP=1),
     kp5, km5)

# Receptor dimer internalization/degradation
Rule('egfr_dimer_degrade',
     EGF(R=1) % EGF(R=2) % EGFR(L=1, CR1=3) % EGFR(L=2, CR1=3) >> None,
     deg)


if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")



from pysb.simulator.base import Simulator, SimulationResult
import scipy.integrate
try:
    # weave is not available under Python 3.
    from weave import inline as weave_inline
    import weave.build_tools
    import distutils.errors
except ImportError:
    weave_inline = None
try:
    import theano.tensor
    from sympy.printing.theanocode import theano_function
except ImportError:
    theano = None
try:
    import Cython
except ImportError:
    Cython = None
from sympy.printing.lambdarepr import lambdarepr
import distutils
import pysb.bng
import sympy
import re
import numpy as np
import warnings
import os
from pysb.logging import get_logger, EXTENDED_DEBUG
import logging
import itertools
import contextlib
import importlib
from concurrent.futures import ProcessPoolExecutor, Executor, Future


class ScipyOdeSimulator(Simulator):
    """
    Simulate a model using SciPy ODE integration

    Uses :func:`scipy.integrate.odeint` for the ``lsoda`` integrator,
    :func:`scipy.integrate.ode` for all other integrators.

    .. warning::
        The interface for this class is considered experimental and may
        change without warning as PySB is updated.

    Parameters
    ----------
    model : pysb.Model
        Model to simulate.
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range. Returned trajectories are sampled at every value unless
        the simulation is interrupted for some reason, e.g., due to
        satisfaction of a logical stopping criterion (see 'tout' below).
    initials : vector-like or dict, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initials (with initial condition
        parameter values taken from `param_values` if specified).
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.
    **kwargs : dict
        Extra keyword arguments, including:

        * ``integrator``: Choice of integrator, including ``vode`` (default),
          ``zvode``, ``lsoda``, ``dopri5`` and ``dop853``. See
          :func:`scipy.integrate.ode` for further information.
        * ``integrator_options``: A dictionary of keyword arguments to
          supply to the integrator. See :func:`scipy.integrate.ode`.
        * ``compiler``: Choice of compiler for ODE system: ``cython``,
          ``weave`` (Python 2 only), ``theano`` or ``python``. Leave
          unspecified or equal to None for auto-select (tries weave,
          then cython, then python). Cython, weave and theano all compile the
          equation system into C code. Python is the slowest but most
          compatible.
        * ``cleanup``: Boolean, `cleanup` argument used for
          :func:`pysb.bng.generate_equations` call

    Notes
    -----
    If ``tspan`` is not defined, it may be defined in the call to the
    ``run`` method.

    Examples
    --------
    Simulate a model and display the results for an observable:

    >>> from pysb.examples.robertson import model
    >>> import numpy as np
    >>> np.set_printoptions(precision=4)
    >>> sim = ScipyOdeSimulator(model, tspan=np.linspace(0, 40, 10))
    >>> simulation_result = sim.run()
    >>> print(simulation_result.observables['A_total']) \
        #doctest: +NORMALIZE_WHITESPACE
    [1.      0.899   0.8506  0.8179  0.793   0.7728  0.7557  0.7408  0.7277
    0.7158]

    For further information on retrieving trajectories (species,
    observables, expressions over time) from the ``simulation_result``
    object returned by :func:`run`, see the examples under the
    :class:`SimulationResult` class.
    """

    _supports = {'multi_initials': True,
                 'multi_param_values': True}

    # some sane default options for a few well-known integrators
    default_integrator_options = {
        'vode': {
            'method': 'bdf',
            'with_jacobian': True,
            # Set nsteps as high as possible to give our users flexibility in
            # choosing their time step. (Let's be safe and assume vode was
            # compiled with 32-bit ints. What would actually happen if it was
            # and we passed 2**64-1 though?)
            'nsteps': 2 ** 31 - 1,
        },
        'cvode': {
            'method': 'bdf',
            'iteration': 'newton',
        },
        'lsoda': {
            'mxstep': 2**31-1,
        }
    }

    default_cython_directives = {
        'boundscheck': False,
        'wraparound': False,
        'nonecheck': False,
        'initializedcheck': False
    }

    def __init__(self, model, tspan=None, initials=None, param_values=None,
                 verbose=False, **kwargs):

        super(ScipyOdeSimulator, self).__init__(model,
                                                tspan=tspan,
                                                initials=initials,
                                                param_values=param_values,
                                                verbose=verbose,
                                                **kwargs)
        # We'll need to know if we're using the Jacobian when we get to run()
        self._use_analytic_jacobian = kwargs.pop('use_analytic_jacobian',
                                                 False)
        self.cleanup = kwargs.pop('cleanup', True)
        integrator = kwargs.pop('integrator', 'vode')
        compiler_mode = kwargs.pop('compiler', None)
        integrator_options = kwargs.pop('integrator_options', {})

        if kwargs:
            raise ValueError('Unknown keyword argument(s): {}'.format(
                ', '.join(kwargs.keys())
            ))
        # Generate the equations for the model
        pysb.bng.generate_equations(self._model, self.cleanup, self.verbose)

        # ODE RHS -----------------------------------------------
        self._eqn_subs = {e: e.expand_expr(expand_observables=True) for
                          e in self._model.expressions}
        self._eqn_subs.update({e: e.expand_expr(expand_observables=True) for
                               e in self._model._derived_expressions})
        ode_mat = sympy.Matrix(self.model.odes).subs(self._eqn_subs)

        if compiler_mode is None:
            self._compiler = self._autoselect_compiler()
            if self._compiler == 'python':
                self._logger.warning(
                    "This system of ODEs will be evaluated in pure Python. "
                    "This may be slow for large models. We recommend "
                    "installing a package for compiling the ODEs to C code: "
                    "'weave' (recommended for Python 2) or "
                    "'cython' (recommended for Python 3). This warning can "
                    "be suppressed by specifying compiler='python'.")
            self._logger.debug('Equation mode set to "%s"' % self._compiler)
        else:
            self._compiler = compiler_mode

        self._compiler_directives = None

        # Use lambdarepr (Python code) with Cython, otherwise use C code
        eqn_repr = lambdarepr if self._compiler == 'cython' else sympy.ccode

        if self._compiler in ('weave', 'cython'):
            # Prepare the string representations of the RHS equations

            code_eqs = '\n'.join(['ydot[%d] = %s;' %
                                  (i, eqn_repr(o))
                                  for i, o in enumerate(ode_mat)])
            code_eqs = str(self._eqn_substitutions(code_eqs))

            # Allocate ydot here, once.
            ydot = np.zeros(len(self.model.species))

            if self._compiler == 'cython':
                self._compiler_directives = kwargs.pop(
                    'cython_directives', self.default_cython_directives
                )

                if not Cython:
                    raise ImportError('Cython library is not installed')

                rhs = _get_rhs(self._compiler,
                               code_eqs,
                               ydot=ydot,
                               compiler_directives=self._compiler_directives
                               )

                with _set_cflags_no_warnings(self._logger):
                    rhs(0.0, self.initials[0], self.param_values[0])
            else:
                # Weave
                self._compiler_directives = []
                # Inhibit weave C compiler warnings unless log
                # level <= EXTENDED_DEBUG. Note that since the output goes
                # straight to stderr rather than via the logging system, the
                # threshold must be lower than DEBUG or else the Nose
                # logcapture plugin will cause the warnings to be shown and
                # tests will fail due to unexpected output.
                if not self._logger.isEnabledFor(EXTENDED_DEBUG):
                    self._compiler_directives.append('-w')
                if not weave_inline:
                    raise ImportError('Weave library is not installed')
                for arr_name in ('ydot', 'y', 'p'):
                    macro = arr_name.upper() + '1'
                    code_eqs = re.sub(r'\b%s\[(\d+)\]' % arr_name,
                                      '%s(\\1)' % macro, code_eqs)

                rhs = _get_rhs(self._compiler,
                               code_eqs,
                               ydot=ydot,
                               compiler_directives=self._compiler_directives
                               )

                # Call rhs once just to trigger the weave C compilation step
                # while asserting control over distutils logging.
                with self._patch_distutils_logging:
                    rhs(0.0, self.initials[0], self.param_values[0])

            self._code_eqs = code_eqs

        elif self._compiler in ('theano', 'python'):
            self._symbols = sympy.symbols(','.join('__s%d' % sp_id for sp_id in
                                                   range(len(
                                                       self.model.species)))
                                          + ',') + tuple(model.parameters)

            if self._compiler == 'theano':
                warnings.warn(
                    "theano backend is deprecated; cython backend is "
                    "recommended instead",
                    category=DeprecationWarning,
                    stacklevel=2
                )
                if theano is None:
                    raise ImportError('Theano library is not installed')

                code_eqs_py = theano_function(
                    self._symbols,
                    [o if not o.is_zero else theano.tensor.zeros(1)
                     for o in ode_mat],
                    on_unused_input='ignore'
                )
            else:
                code_eqs_py = sympy.lambdify(self._symbols,
                                             sympy.flatten(ode_mat))

            rhs = _get_rhs(self._compiler, code_eqs_py)
            self._code_eqs = code_eqs_py
        else:
            raise ValueError('Unknown compiler_mode: %s' % self._compiler)

        # JACOBIAN -----------------------------------------------
        # We'll keep the code for putting together the matrix in Sympy
        # in case we want to do manipulations of the matrix later (e.g., to
        # put together the sensitivity matrix)
        jac_fn = None
        self._jac_eqs = None
        if self._use_analytic_jacobian:
            species_symbols = [sympy.Symbol('__s%d' % i)
                               for i in range(len(self._model.species))]
            jac_matrix = ode_mat.jacobian(species_symbols)

            if self._compiler == 'theano':
                jac_eqs_py = theano_function(
                    self._symbols,
                    [j if not j.is_zero else theano.tensor.zeros(1)
                     for j in jac_matrix],
                    on_unused_input='ignore'
                )

                def jac_fn(t, y, p):
                    jacmat = np.asarray(jac_eqs_py(*itertools.chain(y, p)))
                    jacmat.shape = (len(self.model.odes),
                                    len(self.model.species))
                    return jacmat

            elif self._compiler in ('weave', 'cython'):
                # Prepare the stringified Jacobian equations.
                jac_eqs_list = []
                for i in range(jac_matrix.shape[0]):
                    for j in range(jac_matrix.shape[1]):
                        entry = jac_matrix[i, j]
                        # Skip zero entries in the Jacobian
                        if entry == 0:
                            continue
                        jac_eq_str = 'jac[%d, %d] = %s;' % (
                            i, j, eqn_repr(entry))
                        jac_eqs_list.append(jac_eq_str)
                jac_eqs = str(self._eqn_substitutions('\n'.join(jac_eqs_list)))
                if '# Not supported in Python' in jac_eqs:
                    raise ValueError('Analytic Jacobian calculation failed')

                # Allocate jac array here, once, and initialize to zeros.
                jac = np.zeros(
                    (len(self._model.odes), len(self._model.species)))

                if self._compiler == 'weave':
                    # Substitute array refs with calls to the JAC1 macro
                    jac_eqs = re.sub(r'\bjac\[(\d+), (\d+)\]',
                                     r'JAC2(\1, \2)', jac_eqs)
                    # Substitute calls to the Y1 and P1 macros
                    for arr_name in ('y', 'p'):
                        macro = arr_name.upper() + '1'
                        jac_eqs = re.sub(r'\b%s\[(\d+)\]' % arr_name,
                                         '%s(\\1)' % macro, jac_eqs)

                    jac_fn = _get_rhs(
                        self._compiler,
                        jac_eqs,
                        compiler_directives=self._compiler_directives,
                        jac=jac
                    )

                    # Manage distutils logging, as above for rhs.
                    with self._patch_distutils_logging:
                        jac_fn(0.0, self.initials[0], self.param_values[0])
                else:
                    jac_fn = _get_rhs(
                        self._compiler,
                        jac_eqs,
                        compiler_directives=self._compiler_directives,
                        jac=jac
                    )

                    with _set_cflags_no_warnings(self._logger):
                        jac_fn(0.0, self.initials[0], self.param_values[0])
                self._jac_eqs = jac_eqs
            else:
                jac_eqs_py = sympy.lambdify(self._symbols, jac_matrix, "numpy")

                jac_fn = _get_rhs(self._compiler, jac_eqs_py)

                self._jac_eqs = jac_eqs_py

        # build integrator options list from our defaults and any kwargs
        # passed to this function
        options = {}
        if self.default_integrator_options.get(integrator):
            options.update(
                self.default_integrator_options[integrator])  # default options

        options.update(integrator_options)  # overwrite
        # defaults
        self.opts = options

        if integrator != 'lsoda':
            # Only used to check the user has selected a valid integrator
            self.integrator = scipy.integrate.ode(rhs, jac=jac_fn)
            with warnings.catch_warnings():
                warnings.filterwarnings('error', 'No integrator name match')
                self.integrator.set_integrator(integrator, **options)

    @property
    def _patch_distutils_logging(self):
        """Return distutils logging context manager based on our logger."""
        return _patch_distutils_logging(self._logger.logger)

    @classmethod
    def _test_inline(cls):
        """
        Detect whether weave.inline is functional.

        Produces compile warnings, which we suppress by capturing STDERR.
        """
        if not hasattr(cls, '_use_inline'):
            cls._use_inline = False
            if weave_inline is not None:
                logger = get_logger(__name__)
                extra_compile_args = []
                # See comment in __init__ for why this must be EXTENDED_DEBUG.
                if not logger.isEnabledFor(EXTENDED_DEBUG):
                    if os.name == 'posix':
                        extra_compile_args.append('2>/dev/null')
                    elif os.name == 'nt':
                        extra_compile_args.append('2>NUL')
                try:
                    with _patch_distutils_logging(logger):
                        weave_inline('int i=0; i=i;', force=1,
                                     extra_compile_args=extra_compile_args)
                    cls._use_inline = True
                except (weave.build_tools.CompileError,
                        distutils.errors.CompileError,
                        ImportError,
                        ValueError) as e:
                    if not cls._check_compiler_error(e, 'weave'):
                        raise

    @classmethod
    def _test_cython(cls):
        if not hasattr(cls, '_use_cython'):
            cls._use_cython = False
            if Cython is None:
                return
            try:
                Cython.inline('x = 1', force=True, quiet=True)
                cls._use_cython = True
            except (Cython.Compiler.Errors.CompileError,
                    distutils.errors.DistutilsPlatformError,
                    ValueError) as e:
                if not cls._check_compiler_error(e, 'cython'):
                    raise

    @staticmethod
    def _check_compiler_error(e, compiler):
        if isinstance(e, distutils.errors.DistutilsPlatformError) and \
                str(e) != 'Unable to find vcvarsall.bat':
            return False

        if isinstance(e, ValueError) and e.args != ('Symbol table not found',):
            return False

        # Build platform-specific C compiler error message
        message = 'Please check you have a functional C compiler'
        if os.name == 'nt':
            message += ', available from ' \
                       'https://wiki.python.org/moin/WindowsCompilers'
        else:
            message += '.'

        get_logger(__name__).warn(
            '{} compiler appears non-functional. {}\n'
            'Original error: {}'.format(compiler, message, repr(e)))

        return True

    @classmethod
    def _autoselect_compiler(cls):
        """ Auto-select equation backend """

        # Try weave
        cls._test_inline()
        if cls._use_inline:
            return 'weave'

        # Try cython
        cls._test_cython()
        if cls._use_cython:
            return 'cython'

        # Default to python/lambdify
        return 'python'

    def _eqn_substitutions(self, eqns):
        """String substitutions on the sympy C code for the ODE RHS and
        Jacobian functions to use appropriate terms for variables and
        parameters."""
        # Substitute 'y[i]' for 'si'
        eqns = re.sub(r'\b__s(\d+)\b',
                      lambda m: 'y[%s]' % (int(m.group(1))),
                      eqns)

        # Substitute 'p[i]' for any named parameters
        for i, p in enumerate(self._model.parameters):
            eqns = re.sub(r'\b(%s)\b' % p.name, 'p[%d]' % i, eqns)
        for i, p in enumerate(self._model._derived_parameters):
            eqns = re.sub(r'\b(%s)\b' % p.name,
                          'p[%d]' % (i + len(self._model.parameters)), eqns)
        return eqns

    def run(self, tspan=None, initials=None, param_values=None,
            num_processors=1):
        """
        Run a simulation and returns the result (trajectories)

        .. note::
            In early versions of the Simulator class, ``tspan``, ``initials``
            and ``param_values`` supplied to this method persisted to future
            :func:`run` calls. This is no longer the case.

        Parameters
        ----------
        tspan
        initials
        param_values
            See parameter definitions in :class:`ScipyOdeSimulator`.
        num_processors : int
            Number of processes to use (default: 1). Set to a larger number
            (e.g. the number of CPU cores available) for parallel execution of
            simulations. This is only useful when simulating with more than one
            set of initial conditions and/or parameters.

        Returns
        -------
        A :class:`SimulationResult` object
        """
        super(ScipyOdeSimulator, self).run(tspan=tspan,
                                           initials=initials,
                                           param_values=param_values,
                                           _run_kwargs=[])
        n_sims = len(self.param_values)

        num_species = len(self._model.species)
        num_odes = len(self._model.odes)
        results = []
        if num_processors == 1:
            self._logger.debug('Single processor (serial) mode')
        else:
            self._logger.debug('Multi-processor (parallel) mode using {} '
                               'processes'.format(num_processors))
        with SerialExecutor() if num_processors == 1 else \
                ProcessPoolExecutor(max_workers=num_processors) as executor:
            for n in range(n_sims):
                results.append(executor.submit(
                    _integrator_process,
                    self._code_eqs,
                    self._jac_eqs,
                    num_species,
                    num_odes,
                    self.initials[n],
                    self.tspan,
                    self.param_values[n],
                    self._init_kwargs.get('integrator', 'vode'),
                    compiler=self._compiler,
                    integrator_opts=self.opts,
                    compiler_directives=self._compiler_directives
                ))
            trajectories = [r.result() for r in results]

        tout = np.array([self.tspan] * n_sims)
        self._logger.info('All simulation(s) complete')
        return SimulationResult(self, tout, trajectories)


@contextlib.contextmanager
def _patch_distutils_logging(base_logger):
    """Patch distutils logging functionality with logging.Logger calls.

    The value of the 'base_logger' argument should be a logging.Logger instance,
    and its effective level will be passed on to the patched distutils loggers.

    distutils.log contains its own internal PEP 282 style logging system that
    sends messages straight to stdout/stderr, and numpy.distutils.log extends
    that. This code patches all of this with calls to logging.LoggerAdapter
    instances, and disables the module-level threshold-setting functions so we
    can retain full control over the threshold. Also all WARNING messages are
    "downgraded" to INFO to suppress excessive use of WARNING-level logging in
    numpy.distutils.

    """
    logger = get_logger(__name__)
    logger.debug('patching distutils and numpy.distutils logging')
    logger_methods = 'log', 'debug', 'info', 'warn', 'error', 'fatal'
    other_functions = 'set_threshold', 'set_verbosity'
    saved_symbols = {}
    for module_name in 'distutils.log', 'numpy.distutils.log':
        new_logger = _DistutilsProxyLoggerAdapter(
            base_logger, {'module': module_name}
        )
        module = importlib.import_module(module_name)
        # Save the old values.
        for name in logger_methods + other_functions:
            saved_symbols[module, name] = getattr(module, name)
        # Replace logging functions with bound methods of the Logger object.
        for name in logger_methods:
            setattr(module, name, getattr(new_logger, name))
        # Replace threshold-setting functions with no-ops.
        for name in other_functions:
            setattr(module, name, lambda *args, **kwargs: None)
    try:
        yield
    finally:
        logger.debug('restoring distutils and numpy.distutils logging')
        # Restore everything we overwrote.
        for (module, name), value in saved_symbols.items():
            setattr(module, name, value)


@contextlib.contextmanager
def _set_cflags_no_warnings(logger):
    """ Suppress cython warnings by setting -w flag """
    del_cflags = False
    if 'CFLAGS' not in os.environ \
            and not logger.isEnabledFor(EXTENDED_DEBUG):
        del_cflags = True
        os.environ['CFLAGS'] = '-w'
    try:
        yield
    finally:
        if del_cflags:
            del os.environ['CFLAGS']


class _DistutilsProxyLoggerAdapter(logging.LoggerAdapter):
    """A logging adapter for the distutils logging patcher."""
    def process(self, msg, kwargs):
        return '(from %s) %s' % (self.extra['module'], msg), kwargs
    # Map 'warn' to 'info' to reduce chattiness.
    warn = logging.LoggerAdapter.info
    # Provide 'fatal' to match up with distutils log functions.
    fatal = logging.LoggerAdapter.critical


def _get_rhs(compiler, code_eqs, ydot=None, jac=None, compiler_directives=None):
    if compiler == 'cython':
        if 'math.' in code_eqs:
            code_eqs = 'import math\n' + code_eqs

        def rhs(t, y, p):
            # note that the evaluated code sets ydot as a side effect
            Cython.inline(code_eqs, quiet=True,
                          cython_compiler_directives=compiler_directives)

            return ydot if ydot is not None else jac
    elif compiler == 'weave':
        def rhs(t, y, p):
            # note that the evaluated code sets ydot as a side effect
            weave_inline(code_eqs, ['ydot' if ydot is not None else 'jac',
                                    't', 'y', 'p'],
                         extra_compile_args=compiler_directives)
            return ydot if ydot is not None else jac
    else:
        def rhs(t, y, p):
            return code_eqs(*itertools.chain(y, p))

    return rhs


def _integrator_process(code_eqs, jac_eqs, num_species, num_odes, initials,
                        tspan, param_values, integrator_name, compiler,
                        integrator_opts, compiler_directives):
    """ Single integrator process, for parallel execution """
    rhs = _get_rhs(compiler, code_eqs, ydot=np.zeros(num_species),
                   compiler_directives=compiler_directives)

    jac_fn = None
    if jac_eqs:
        jac_eqs = _get_rhs(compiler, jac_eqs,
                           jac=np.zeros((num_odes, num_species)),
                           compiler_directives=compiler_directives)

    # LSODA
    if integrator_name == 'lsoda':
        return scipy.integrate.odeint(
            rhs,
            initials,
            tspan,
            args=(param_values, ),
            Dfun=jac_fn,
            tfirst=True,
            **integrator_opts
        )

    # All other integrators
    integrator = scipy.integrate.ode(rhs, jac=jac_fn)
    with warnings.catch_warnings():
        warnings.filterwarnings('error', 'No integrator name match')
        integrator.set_integrator(integrator_name, **integrator_opts)
    integrator.set_initial_value(initials, tspan[0])

    # Set parameter vectors for RHS func and Jacobian
    integrator.set_f_params(param_values)
    if jac_eqs:
        integrator.set_jac_params(param_values)

    trajectory = np.ndarray((len(tspan), num_species))
    trajectory[0] = initials
    i = 1
    while integrator.successful() and integrator.t < tspan[-1]:
        trajectory[i] = integrator.integrate(tspan[i])
        i += 1
    if integrator.t < tspan[-1]:
        trajectory[i:, :] = 'nan'

    return trajectory


class SerialExecutor(Executor):
    """ Execute tasks in serial (immediately on submission) """
    def submit(self, fn, *args, **kwargs):
        f = Future()
        try:
            result = fn(*args, **kwargs)
        except BaseException as e:
            f.set_exception(e)
        else:
            f.set_result(result)

        return f



# -*- coding:utf-8; -*-
"""
Test example models export without error using the available exporters

Based on code submitted in a PR by @keszybz in pysb/pysb#113
"""
from pysb.tests.test_examples import get_example_models, expected_exceptions
from pysb import export
from pysb.simulator import ScipyOdeSimulator
import numpy as np
import pandas as pd
try:
    import roadrunner
except ImportError:
    roadrunner = None
from nose.plugins.skip import SkipTest


# Pairs of model, format that are expected to be incompatible.
skip_combinations = {
    ('fixed_initial', 'kappa'),
}


def test_export():
    for model in get_example_models():
        for format in export.formats:
            if (base_name(model), format) in skip_combinations:
                continue
            fn = lambda: check_convert(model, format)
            fn.description = "Check export: {} → {}".format(
                model.name, format)
            yield fn


def base_name(model):
    # This needs to handle names with zero or more dots.
    return model.name.split('.')[-1]


def check_convert(model, format):
    """ Test exporters run without error """
    exported_file = None
    try:
        exported_file = export.export(model, format)
    except export.ExpressionsNotSupported:
        pass
    except export.CompartmentsNotSupported:
        pass
    except export.LocalFunctionsNotSupported:
        pass
    except Exception as e:
        # Some example models are deliberately incomplete, so here we
        # will treat any of these "expected" exceptions as a success.
        exception_class = expected_exceptions.get(base_name(model))
        if not exception_class or not isinstance(e, exception_class):
            raise

    if exported_file is not None:
        if format == 'python':
            # linspace arguments picked to avoid VODE warning
            exec(exported_file + 'Model().simulate(tspan=numpy.linspace(0,1,501))\n', {'_use_inline': False})
        elif format == 'pysb_flat':
            exec(exported_file, {'__name__': model.name})
        elif format == 'sbml':
            # Skip the simulation comparison if roadrunner not available
            if roadrunner is None:
                raise SkipTest("SBML Simulation test skipped (requires roadrunner)")

            roadrunner.Logger.setLevel(roadrunner.Logger.LOG_ERROR)

            # Simulate SBML using roadrunner
            rr = roadrunner.RoadRunner(exported_file)
            rr.timeCourseSelections = \
                ['__s{}'.format(i) for i in range(len(model.species))] + \
                ['__obs{}'.format(i) for i in range(len(model.observables))]
            rr_result = rr.simulate(0, 10, 100)

            # Simulate original using PySB
            df = ScipyOdeSimulator(model).run(tspan=np.linspace(0, 10, 100)).dataframe

            # Compare species' trajectories
            for sp_idx in range(len(model.species)):
                rr_sp = rr_result[:, sp_idx]
                py_sp = df.iloc[:, sp_idx]
                is_close = np.allclose(rr_sp, py_sp, rtol=1e-4)
                if not is_close:
                    print(pd.DataFrame(dict(rr=rr_sp, pysb=py_sp)))
                    raise ValueError('Model {}, species __s{} trajectories do not match:'.format(
                        model.name, sp_idx))

            # Compare observables' trajectories
            for obs_idx in range(len(model.observables)):
                rr_obs = rr_result[:, obs_idx + len(model.species)]
                py_obs = df.iloc[:, obs_idx + len(model.species)]
                is_close = np.allclose(rr_obs, py_obs, rtol=1e-4)
                if not is_close:
                    print(pd.DataFrame(dict(rr=rr_obs, pysb=py_obs)))
                    raise ValueError('Model {}, observable__o{} "{}" trajectories do not match:'.format(
                        model.name, obs_idx, model.observables[obs_idx].name))



"""EARM 1.3 (extrinsic apoptosis reaction model)

Gaudet S, Spencer SL, Chen WW, Sorger PK (2012) Exploring the Contextual
Sensitivity of Factors that Determine Cell-to-Cell Variability in
Receptor-Mediated Apoptosis. PLoS Comput Biol 8(4): e1002482.
doi:10.1371/journal.pcbi.1002482

http://www.ploscompbiol.org/article/info:doi/10.1371/journal.pcbi.1002482
"""

# This model is mostly an extension to EARM 1.0 from Albeck et al. Instead of
# duplicating the body of that model here, we'll start this model off as a copy
# of that one then modify and add components as necessary.

from __future__ import print_function
import re
from pysb import *
import pysb.bng
import pysb.examples.earm_1_0

# Start from EARM 1.0 as a base
# ==========
Model(base=pysb.examples.earm_1_0.model)
Annotation(model, 'http://identifiers.org/doi/10.1371/journal.pcbi.1002482',
           'isDescribedBy')

# Rename all instances of Bcl2c to Mcl1, which was determined as a more accurate
# description of that monomer based on its role in the model
# ==========

# monomer
Bcl2c.rename('Mcl1')
# initial condition parameter
Bcl2c_0.rename('Mcl1_0')
# rule
inhibit_tBid_by_Bcl2c.rename('inhibit_tBid_by_Mcl1')

# Add one novel reaction
# ==========

# degrade DISC
Parameter('kf31', 0.001)
Rule('DISC_deg', DISC(b=None) >> L(b=None) + pR(b=None), kf31)
# NOTE: In the original model this is a reversible reaction with kr31 as its
# reverse rate constant, but kr31 was set to 0. This was ostensibly done so all
# reactions had the same symmetric form, but I see no reason not to make it
# irreversible here and eliminate kr31. -JLM

# Change some initial condition values
# ==========

pR_0.value = 1000
flip_0.value = 2000
pC8_0.value = 10000
Bid_0.value = 60000
Bax_0.value = 80000
Bcl2_0.value = 30000

# Change some rate constants
# ==========

v = 0.01

kr1.value = 1e-6
kc1.value = 1e-2
kf3.value = 1e-7
kf6.value = 1e-7
kf7.value = 1e-7
kr9.value = 0.001
kc9.value = 20
kr13.value = 1
kf14.value = 1e-6 / v
kf15.value = 1e-6 * 2 / v
kf16.value = 1e-6 / v
kf17.value = 1e-6 * 2 / v
kf18.value = 1e-6 / v
kf19.value = 1e-6 / v
kf20.value = 2e-6 / v
kf21.value = 2e-6 / v
kf22.value = 1
kf26.value = 1

# Add synthesis and degradation rules and parameters
# ==========

# generate species list and make a copy
pysb.bng.generate_equations(model)
all_species = list(model.species)
model.reset_equations()

def synthesize(name, species, ks):
    """Synthesize species with rate ks"""
    Rule(name, None >> species, ks)

def degrade(name, species, kdeg):
    """Degrade species with rate kdeg"""
    Rule(name, species >> None, kdeg)

# almost all degradation rates use this one value
kdeg_generic = 2.9e-6
# and these are the three exceptions
Parameter('kdeg_Mcl1', 0.0001)
Parameter('kdeg_AMito', 0.0001)
Parameter('kdeg_C3_U', 0)
# fraction by which synthesis rates should be scaled (to mimic the effects of
# treating HeLa cells wth 2.5 ug/ml of cycloheximide as per Table S2, Note 2)
syn_base = 0.15
# synthesis rates are all syn_base*kdeg*IC, except for L which is 0
Parameter('ks_L', 0)

# Even though the degradation of AMito is counted as a degradation reaction, it
# is different from the others in that the reactant species is not destroyed but
# rather converted to another species. So we have to declare this one
# explicitly.
Rule('AMito_deg', AMito(b=None) >> Mito(b=None), kdeg_AMito)

# loop over all species and create associated synthesis and degradation rates
# and reactions
for species in all_species:
    species_name = '_'.join(mp.monomer.name for mp in species.monomer_patterns)
    ks_name = 'ks_' + species_name
    kdeg_name = 'kdeg_' + species_name
    syn_rule_name = species_name + '_syn'
    deg_rule_name = species_name + '_deg'
    ic_name = species_name + '_0'
    ks = model.parameters.get(ks_name)
    kdeg = model.parameters.get(kdeg_name)
    syn_rule = model.rules.get(syn_rule_name)
    deg_rule = model.rules.get(deg_rule_name)
    ic = model.parameters.get(ic_name)
    if kdeg is None:
        kdeg = Parameter(kdeg_name, kdeg_generic)
    if ks is None:
        ks_value = 0
        if ic is not None:
            ks_value = syn_base * kdeg.value * ic.value
        ks = Parameter(ks_name, ks_value)
    if syn_rule is None:
        synthesize(syn_rule_name, species, ks)
    if deg_rule is None:
        degrade(deg_rule_name, species, kdeg)


# ========================================


def show_species():
    """Print a table of species like Table S2"""
    print('   | %-12s %8s %20s %10s' %
          ('species', 'initial', 'synth rate', 'deg rate'))
    print('-' * (5 + 12 + 1 + 8 + 1 + 20 + 1 + 10))
    for i, species in enumerate(all_species, 1):
        mp_names = [mp.monomer.name for mp in species.monomer_patterns]
        name = '_'.join(mp_names)
        display_name = ':'.join(mp_names)
        ks = model.parameters.get('ks_' + name)
        kdeg = model.parameters.get('kdeg_' + name)
        ic = model.parameters.get(name + '_0')
        if ic is not None:
            ic_value = ic.value
        else:
            ic_value = 0
        if ks.value != 0 and kdeg.value != 0:
            ic_calc = round(ks.value/kdeg.value/syn_base)
            ks_expr = '%4.2f*kdeg*%7d' % (syn_base, ic_calc)
        else:
            ks_expr = '0'
        values = (i, display_name, ic_value, ks_expr, kdeg.value)
        print('%2d | %-12s %8d %20s %10g' % values)

def show_rates():
    """Print a table of rate parameters like Table S4"""
    # FIXME kf14-kf21 need to be un-scaled by v to make the table look right
    print(("%-20s        " * 3) % ('forward', 'reverse', 'catalytic'))
    print('-' * (9 + 11 + 7) * 3)
    for i in list(range(1,29)) + [31]:
        for t in ('f', 'r', 'c'):
            n = 'k%s%d' % (t,i)
            p = model.parameters.get(n)
            if p is not None:
                print("%-9s%11g       " % (p.name, p.value), end=' ')
        print()


if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
Initial species concentrations and synthesis/degradation rates
==========
(compare to supporting text S3, table S2 from Gaudet et al.,
however note that the numbering and ordering is different.)""")
    show_species()
    print("\n")

    print("""
Kinetic rate parameters
==========
(compare to supporting text S3, table S4 from Gaudet et al.,
however note that kf14-kf21 are off by a factor of 'v' (%g) as
this volume scaling factor has been moved from the ODEs to the
rate values in this implementation)
""" % v)
    show_rates()
    print("""

NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")



from pysb.simulator import ScipyOdeSimulator, BngSimulator
from pysb.simulator.base import SimulationResult
from pysb.examples import tyson_oscillator, robertson, \
    expression_observables, earm_1_3, bax_pore_sequential, bax_pore, \
    bngwiki_egfr_simple
from pysb.bng import generate_equations
import numpy as np
import tempfile
from nose.tools import assert_raises, raises
import warnings
from pysb.pattern import SpeciesPatternMatcher
import collections
import copy
import pandas as pd


def test_simres_dataframe():
    """ Test SimulationResult.dataframe() """

    tspan1 = np.linspace(0, 100, 100)
    tspan2 = np.linspace(50, 100, 50)
    tspan3 = np.linspace(100, 150, 100)
    model = tyson_oscillator.model
    sim = ScipyOdeSimulator(model, integrator='lsoda')
    simres1 = sim.run(tspan=tspan1)
    # Check retrieving a single simulation dataframe
    df_single = simres1.dataframe

    # Generate multiple trajectories
    trajectories1 = simres1.species
    trajectories2 = sim.run(tspan=tspan2).species
    trajectories3 = sim.run(tspan=tspan3).species

    # Try a simulation result with two different tspan lengths
    sim = ScipyOdeSimulator(model, param_values={'k6' : [1.,1.]}, integrator='lsoda')
    simres = SimulationResult(sim, [tspan1, tspan2], [trajectories1, trajectories2])
    df = simres.dataframe

    assert df.shape == (len(tspan1) + len(tspan2),
                        len(model.species) + len(model.observables))

    # Next try a simulation result with two identical tspan lengths, stacked
    # into a single 3D array of trajectories
    simres2 = SimulationResult(sim, [tspan1, tspan3],
                               np.stack([trajectories1, trajectories3]))
    df2 = simres2.dataframe

    assert df2.shape == (len(tspan1) + len(tspan3),
                         len(model.species) + len(model.observables))


def test_simres_observable():
    """ Test on demand observable evaluation """
    models = [tyson_oscillator.model, robertson.model,
              expression_observables.model, earm_1_3.model,
              bax_pore_sequential.model, bax_pore.model,
              bngwiki_egfr_simple.model]
    for model in models:
        generate_equations(model)
        spm = SpeciesPatternMatcher(model)
        for obs in model.observables:
            dyn_obs = spm.match(pattern=obs.reaction_pattern, index=True,
                                counts=True)

            # Need to sort by species numerical order for comparison purposes
            dyn_obs = collections.OrderedDict(sorted(dyn_obs.items()))

            dyn_species = list(dyn_obs.keys())

            if obs.match == 'species':
                dyn_coeffs = [1] * len(dyn_obs)
            else:
                dyn_coeffs = list(dyn_obs.values())

            assert dyn_species == obs.species
            assert dyn_coeffs == obs.coefficients


class TestSimulationResultEarm13(object):
    def setUp(self):
        self.model = earm_1_3.model
        self.tspan = np.linspace(0, 100, 101)
        self.sim = ScipyOdeSimulator(self.model, tspan=self.tspan)
        self.simres = self.sim.run()

    @raises(ValueError)
    def test_dynamic_observable_nonpattern(self):
        self.simres.observable('cSmac')

    @raises(ValueError)
    def test_match_nonexistent_pattern(self):
        m = self.model.monomers
        self.simres.observable(m.cSmac() % m.Bid())

    def test_on_demand_observable(self):
        m = self.model.monomers
        assert isinstance(self.simres.observable(m.cSmac()), pd.Series)


def test_save_load():
    tspan = np.linspace(0, 100, 101)
    # Make a copy of model so other tests etc. don't see the changed name.
    model = copy.deepcopy(tyson_oscillator.model)
    test_unicode_name = u'Hello \u2603 and \U0001f4a9!'
    model.name = test_unicode_name
    sim = ScipyOdeSimulator(model, integrator='lsoda')
    simres = sim.run(tspan=tspan, param_values={'k6': 1.0})

    sim_rob = ScipyOdeSimulator(robertson.model, integrator='lsoda')
    simres_rob = sim_rob.run(tspan=tspan)

    # Reset equations from any previous network generation
    robertson.model.reset_equations()
    A = robertson.model.monomers['A']

    # NFsim without expressions
    nfsim1 = BngSimulator(robertson.model)
    nfres1 = nfsim1.run(n_runs=2, method='nf', tspan=np.linspace(0, 1))
    # Test attribute saving (text, float, list)
    nfres1.custom_attrs['note'] = 'NFsim without expressions'
    nfres1.custom_attrs['pi'] = 3.14
    nfres1.custom_attrs['some_list'] = [1, 2, 3]

    # NFsim with expressions
    nfsim2 = BngSimulator(expression_observables.model)
    nfres2 = nfsim2.run(n_runs=1, method='nf', tspan=np.linspace(0, 100, 11))

    with tempfile.NamedTemporaryFile() as tf:
        # Cannot have two file handles on Windows
        tf.close()

        simres.save(tf.name, dataset_name='test', append=True)

        # Try to reload when file contains only one dataset and group
        SimulationResult.load(tf.name)

        simres.save(tf.name, append=True)

        # Trying to overwrite an existing dataset gives a ValueError
        assert_raises(ValueError, simres.save, tf.name, append=True)

        # Trying to write to an existing file without append gives an IOError
        assert_raises(IOError, simres.save, tf.name)

        # Trying to write a SimulationResult to the same group with a
        # different model name results in a ValueError
        assert_raises(ValueError, simres_rob.save, tf.name,
                      dataset_name='robertson', group_name=model.name,
                      append=True)

        simres_rob.save(tf.name, append=True)

        # Trying to load from a file with more than one group without
        # specifying group_name should raise a ValueError
        assert_raises(ValueError, SimulationResult.load, tf.name)

        # Trying to load from a group with more than one dataset without
        # specifying a dataset_name should raise a ValueError
        assert_raises(ValueError, SimulationResult.load, tf.name,
                      group_name=model.name)

        # Load should succeed when specifying group_name and dataset_name
        simres_load = SimulationResult.load(tf.name, group_name=model.name,
                                            dataset_name='test')
        assert simres_load._model.name == test_unicode_name

        # Saving network free results requires include_obs_exprs=True,
        # otherwise a warning should be raised
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            nfres1.save(tf.name, dataset_name='nfsim_no_obs', append=True)
            assert len(w) == 1
            assert issubclass(w[-1].category, UserWarning)

        nfres1.save(tf.name, include_obs_exprs=True,
                    dataset_name='nfsim test', append=True)

        # NFsim load
        nfres1_load = SimulationResult.load(tf.name,
                                            group_name=nfres1._model.name,
                                            dataset_name='nfsim test')

        # NFsim with expression
        nfres2.save(tf.name, include_obs_exprs=True, append=True)
        nfres2_load = SimulationResult.load(tf.name,
                                            group_name=nfres2._model.name)

    _check_resultsets_equal(simres, simres_load)
    _check_resultsets_equal(nfres1, nfres1_load)
    _check_resultsets_equal(nfres2, nfres2_load)


def _check_resultsets_equal(res1, res2):
    try:
        assert np.allclose(res1.species, res2.species)
    except ValueError:
        # Network free simulations don't have species
        pass
    assert np.allclose(res1.tout, res2.tout)
    assert np.allclose(res1.param_values, res2.param_values)
    
    if isinstance(res1.initials, np.ndarray):
        assert np.allclose(res1.initials, res2.initials)
    else:
        for k, v in res1.initials.items():
            assert np.allclose(res1.initials[k], v)

    assert np.allclose(res1._yobs_view, res1._yobs_view)
    if res1._model.expressions_dynamic():
        assert np.allclose(res1._yexpr_view, res2._yexpr_view)

    assert res1.squeeze == res2.squeeze
    assert res1.simulator_class == res2.simulator_class
    assert res1.init_kwargs == res2.init_kwargs
    assert res1.run_kwargs == res2.run_kwargs
    assert res1.n_sims_per_parameter_set == \
           res2.n_sims_per_parameter_set
    assert res1._model.name == res2._model.name
    assert res1.timestamp == res2.timestamp
    assert res1.pysb_version == res2.pysb_version
    assert res1.custom_attrs == res2.custom_attrs



"""
Module containing a class for exporting a PySB model to BNGL.

Serves as a wrapper around :py:class:`pysb.generator.bng.BngGenerator`.

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""

from pysb.generator.bng import BngGenerator
from pysb.export import Exporter

class BnglExporter(Exporter):
    """A class for returning the BNGL for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """

    def export(self):
        """Generate the corresponding BNGL for the PySB model associated
        with the exporter. A wrapper around ``pysb.generator.bng.BngGenerator``.

        Returns
        -------
        string
            The BNGL output for the model.
        """
        bngl_str = ''
        if self.docstring:
            bngl_str += '# ' + self.docstring.replace('\n', '\n# ') + '\n'
        gen = BngGenerator(self.model)
        bngl_str += gen.get_content()
        return bngl_str



import pysb
import sympy
import warnings
from sympy.printing import StrPrinter
from sympy.core import S
import collections
import re
import pysb.logging
from pysb.export import CompartmentsNotSupported, LocalFunctionsNotSupported
# Alias basestring under Python 3 for forwards compatibility
try:
    basestring
except NameError:
    basestring = str


class KappaGenerator(object):

    # Dialect can be either 'complx' or 'kasim' (default)
    def __init__(self, model, dialect='kasim', _warn_no_ic=True,
                 _exclude_ic_param=False):
        if model:
            if model.compartments:
                raise CompartmentsNotSupported()
            if model.tags:
                raise LocalFunctionsNotSupported()
        self.model = model
        self.__content = None
        self.dialect = dialect
        self._warn_no_ic = _warn_no_ic
        self._exclude_ic_param = _exclude_ic_param
        self._renamed_states = collections.defaultdict(dict)
        self._log = pysb.logging.get_logger(__name__)

    def get_content(self):
        if self.__content == None:
            self.generate_content()
        return self.__content

    def generate_content(self):
        self.__content = ''
        #self.generate_compartments()

        # Agent declarations appear to be required in kasim
        # but prohibited in complx
        if (self.dialect == 'kasim'):
            self.generate_molecule_types() 
            # Parameters, variables, and expressions are allowed in kasim
            if not self._exclude_ic_param:
                self.generate_parameters()

        self.generate_reaction_rules()
        self.generate_observables()
        if not self._exclude_ic_param:
            self.generate_species()

    def generate_parameters(self):
        for p in self.model.parameters:
            self.__content += "%%var: '%s' %e\n" % (p.name, p.value)
        for e in self.model.expressions:
            str_expr = str(expression_to_muparser(e))
            self.__content += "%%var: '%s' %s\n" % (e.name, str_expr)
        self.__content += "\n"

    def generate_molecule_types(self):
        for m in self.model.monomers:
            site_code = ','.join([
                self.format_monomer_site(m, s)
                for s in m.sites])
            self.__content += "%%agent: %s(%s)\n" % (m.name, site_code)
        self.__content += "\n"

    def generate_reaction_rules(self):
        for r in self.model.rules:
            label = "'" + r.name + "'"
            reactants_code = self.format_reactionpattern(r.reactant_pattern)
            products_code = self.format_reactionpattern(r.product_pattern)
            arrow = '->'

            if isinstance(r.rate_forward,pysb.core.Expression) and \
               self.dialect == 'complx':
                raise KappaException("Expressions are not supported by complx.")
            # Get the rate code depending on the dialect
            if self.dialect == 'kasim':
                f_rate_code = "'" + r.rate_forward.name + "'"
            else:
                f_rate_code = float(r.rate_forward.value)

            self.__content += ("%s %s %s %s @ %s") % \
                (label, reactants_code, arrow, products_code, f_rate_code)
            self.__content += "\n"

            # Add the reverse reaction
            if r.is_reversible:
                if isinstance(r.rate_reverse,pysb.core.Expression) and \
                   self.dialect == 'complx':
                    raise KappaException("Expressions are not supported by "
                                         "complx.")
                # Get the rate code depending on the dialect
                if self.dialect == 'kasim':
                    r_rate_code = "'" + r.rate_reverse.name + "'"
                else:
                    r_rate_code = float(r.rate_reverse.value)

                label = "'" + r.name + '_rev' + "'"
                self.__content += ("%s %s %s %s @ %s") % \
                      (label, products_code, arrow, reactants_code, r_rate_code)
                self.__content += "\n"

        self.__content += "\n"

    def generate_observables(self):
        if not self.model.observables:
            return
        for obs in self.model.observables:
            name = "'" + obs.name + "'"
            observable_code = self.format_reactionpattern(obs.reaction_pattern)
            self.__content += "%%obs: %s |%s|\n" % (name, observable_code)

        self.__content += "\n"

    def generate_species(self):
        if self._warn_no_ic and not self.model.initials:
            warnings.warn("Warning: No initial conditions.")
        if any(ic.fixed for ic in self.model.initials):
            raise KappaException(
                "Kappa generator does not support fixed-amount species"
            )

        species_codes = [
            self.format_complexpattern(ic.pattern) for ic in self.model.initials
        ]
        #max_length = max(len(code) for code in species_codes)
        for i, code in enumerate(species_codes):
            param = self.model.initials[i].value
            #self.__content += ("%%init:  %-" + str(max_length) + \
            #                  "s   %s\n") % (code, param.name)
            if (self.dialect == 'kasim'):
                self.__content += ("%%init: '%s' %s\n") % (param.name, code)
            else:
                if isinstance(param,pysb.core.Expression):
                    raise KappaException("complx does not support Expressions.")
                # Switched from %g (float) to %d (int) because kappa didn't
                # like scientific notation for large integers
                self.__content += "%%init: %10d * %s\n" % (param.value, code)
        self.__content += "\n"

    def format_monomer_site(self, monomer, site):
        ret = site
        if site in monomer.site_states:
            new_states = []
            for state in monomer.site_states[site]:
                if re.match('_*[0-9]*$', state):
                    new_state = '__state' + state
                    while new_state in new_states:
                        new_state = '_' + new_state
                    self._renamed_states[monomer.name][state] = new_state
                    state = new_state
                new_states.append(state)

            ret += '{%s}' % ' '.join(new_states)
        if self._renamed_states[monomer.name]:
            self._log.info('Monomer "{}" states were renamed as '
                           'follows: {}'.format(
                monomer.name,
                self._renamed_states[monomer.name])
            )
        return ret

    def format_reactionpattern(self, rp):
        if not rp.complex_patterns:
            return '.'
        return ','.join(filter(None,
                               [self.format_complexpattern(cp)
                                for cp in rp.complex_patterns]))

    def format_complexpattern(self, cp):
        if cp is None:
            return '.'
        ret = ','.join(filter(None, [self.format_monomerpattern(mp)
                                     for mp in cp.monomer_patterns]))
        if cp.compartment is not None:
            ret = '@%s:%s' % (cp.compartment.name, ret)
        return ret

    def format_monomerpattern(self, mp):
        # sort sites in the same order given in the original Monomer
        site_conditions = sorted(mp.site_conditions.items(),
                                 key=lambda x: mp.monomer.sites.index(x[0]))
        site_pattern_code = ','.join([self.format_site_condition(
                                        site, state, mp.monomer.name)
                                      for (site, state) in site_conditions])
        ret = '%s(%s)' % (mp.monomer.name, site_pattern_code)
        if mp.compartment is not None:
            ret = '%s@%s' % (ret, mp.compartment.name)
        return ret

    def format_site_condition(self, site, state, mon_name):
        state = self._renamed_states.get(mon_name, {}).get(state, state)
        # If the state/bond is unspecified
        if state is None:
            state_code = '[.]'
        # If there is a bond number
        elif isinstance(state, int):
            state_code = '[%s]' % state
        # Multi-bond (list of bonds)
        elif isinstance(state, list):
            raise KappaException("Kappa generator does not support multi-bonds")
        # If there is a MultiState, raise an Exception (not supported by Kappa)
        elif isinstance(state, pysb.MultiState):
            raise KappaException("Kappa generator does not support MultiStates")
        # Site with state
        elif isinstance(state, basestring):
            state_code = '{%s}[.]' % state
        # Site with state and a bond
        elif isinstance(state, tuple):
            # If the bond is ANY
            if state[1] == pysb.ANY:
                state_code = '{%s}[_]' % state[0]
            # If the bond is WILD
            elif state[1] == pysb.WILD:
                state_code = '{%s}[#]' % state[0]
            # If the bond is a number
            elif type(state[1]) == int:
                state_code = '{%s}[%s]' % state
            # If it's something else, raise an Exception
            else:
                raise KappaException("Kappa generator has encountered an "
                                     "unknown element in a site state/bond "
                                     "tuple: (%s, %s)" % state)
        # Site bound to ANY
        elif state == pysb.ANY:
            state_code = '[_]'
        # Site bound to WILD
        elif state == pysb.WILD:
            state_code = '[#]'
        # Something else
        else:
            raise KappaException("Kappa generator has encountered an unknown "
                                 "element in a rule pattern site condition.")
        return '%s%s' % (site, state_code)


# Old, stateless versions of these functions kept for API compatibility
def format_monomer_site(monomer, site):
    return KappaGenerator(None).format_monomer_site(monomer, site)


def format_reactionpattern(rp):
    return KappaGenerator(None).format_reactionpattern(rp)


def format_complexpattern(cp):
    return KappaGenerator(None).format_complexpattern(cp)


def format_monomerpattern(mp):
    return KappaGenerator(None).format_monomerpattern(mp)


def format_site_condition(site, state):
    return KappaGenerator(None).format_site_condition(site, state, '')
# End API compatibility functions


class KappaPrinter(StrPrinter):
    def __init__(self, **settings):
        super(KappaPrinter, self).__init__(settings)

    def _print_Piecewise(self, expr):
        if expr.args[-1][1] is not sympy.true:
            raise NotImplementedError('Piecewise statements are only '
                                      'supported if convertible to Kappa '
                                      'ternary statements')

        if_stmt = self._print(expr.args[-1][0])
        for pos in range(len(expr.args) - 2, -1, -1):
            if_stmt = '{} [?] {} [:] {}'.format(
                self._print(expr.args[pos][1]),
                self._print(expr.args[pos][0]),
                if_stmt
            )

        return if_stmt

    def _print_Pow(self, expr, rational=False):
        """ Adapted from sympy/printing/str.py """
        if expr.exp is S.Half and not rational:
            return "[sqrt](%s)" % self._print(expr.base)

        if expr.is_commutative:
            if -expr.exp is S.Half and not rational:
                # Note: Don't test "expr.exp == -S.Half" here, because that
                # will match -0.5, which we don't want.
                return "1/[sqrt](%s)" % self._print(expr.base)

        return super(KappaPrinter, self)._print_Pow(expr, rational)\
            .replace('**', '^')

    def _print_Pi(self, expr):
        return '[pi]'

    def _print_Exp1(self, expr):
        return '[E]'

    def _print_Function(self, expr):
        """ Adapted from sympy/printing/str.py """
        if (expr.func in (sympy.log, sympy.exp, sympy.sin, sympy.cos,
                         sympy.tan) and len(expr.args) == 1):
            func_name = '[{}]'.format(expr.func.__name__)
        elif expr.func is sympy.Mod and len(expr.args) == 2:
            return '{} [mod] {}'.format(*map(self._print, expr.args))
        else:
            func_name = expr.func.__name__
        return "{}({})".format(func_name, self.stringify(expr.args, ", "))

    def _print_Max(self, expr):
        """ Adapted from sympy/printing/cxxcode.py """
        if len(expr.args) == 1:
            return self._print(expr.args[0])
        return "[max] {} {}".format(
            self._print(expr.args[0]), self._print(sympy.Max(*expr.args[1:])))

    def _print_Min(self, expr):
        """ Adapted from sympy/printing/cxxcode.py """
        if len(expr.args) == 1:
            return self._print(expr.args[0])
        return "[min] {} {}".format(
            self._print(expr.args[0]), self._print(sympy.Min(*expr.args[1:])))

    def _print_Parameter(self, expr):
        return "'{}'".format(expr.name)


def expression_to_muparser(expression):
    """Render the Expression as a muparser-compatible string."""

    return KappaPrinter(order='none').doprint(expression.expr)


class KappaException(Exception):
    pass



"""
Module containing a class for converting a PySB model to an equivalent set of
ordinary differential equations for integration or analysis in PottersWheel_.

.. _PottersWheel: http://www.potterswheel.de

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.

Output for the Robertson example model
======================================

The PottersWheel code produced will follow the form as given below for
``pysb.examples.robertson``::

    % A simple three-species chemical kinetics system known as "Robertson's
    % example", as presented in:
    % 
    % H. H. Robertson, The solution of a set of reaction rate equations, in Numerical
    % Analysis: An Introduction, J. Walsh, ed., Academic Press, 1966, pp. 178-182.
    % 
    % PottersWheel model definition file
    % save as robertson.m
    function m = robertson()

    m = pwGetEmptyModel();

    % meta information
    m.ID          = 'robertson';
    m.name        = 'robertson';
    m.description = '';
    m.authors     = {''};
    m.dates       = {''};
    m.type        = 'PW-1-5';

    % dynamic variables
    m = pwAddX(m, 's0', 1.000000e+00);
    m = pwAddX(m, 's1', 0.000000e+00);
    m = pwAddX(m, 's2', 0.000000e+00);

    % dynamic parameters
    m = pwAddK(m, 'k1', 4.000000e-02);
    m = pwAddK(m, 'k2', 3.000000e+07);
    m = pwAddK(m, 'k3', 1.000000e+04);
    m = pwAddK(m, 'A_0', 1.000000e+00);
    m = pwAddK(m, 'B_0', 0.000000e+00);
    m = pwAddK(m, 'C_0', 0.000000e+00);

    % ODEs
    m = pwAddODE(m, 's0', '-k1*s0 + k3*s1*s2');
    m = pwAddODE(m, 's1', 'k1*s0 - k2*power(s1, 2) - k3*s1*s2');
    m = pwAddODE(m, 's2', 'k2*power(s1, 2)');

    % observables
    m = pwAddY(m, 'A_total', '1.000000 * s0');
    m = pwAddY(m, 'B_total', '1.000000 * s1');
    m = pwAddY(m, 'C_total', '1.000000 * s2');

    % end of PottersWheel model robertson
"""

import pysb
import pysb.bng
import sympy
import re
import sys
import os
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
from pysb.export import Exporter, ExpressionsNotSupported, \
    CompartmentsNotSupported


class PottersWheelExporter(Exporter):
    """A class for returning the PottersWheel equivalent for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """

    def export(self):
        """Generate the PottersWheel code for the ODEs of the PySB model
        associated with the exporter.

        Returns
        -------
        string
            String containing the PottersWheel code for the ODEs.
        """
        if self.model.expressions:
            raise ExpressionsNotSupported()
        if self.model.compartments:
            raise CompartmentsNotSupported()

        output = StringIO()
        pysb.bng.generate_equations(self.model)

        model_name = self.model.name.replace('.', '_')

        ic_values = [0] * len(self.model.odes)
        for ic in self.model.initials:
            ic_values[self.model.get_species_index(ic.pattern)] = ic.value.value

        # list of "dynamic variables"
        pw_x = ["m = pwAddX(m, 's%d', %e);" % (i, ic_values[i])
                for i in range(len(self.model.odes))]

        # parameters
        pw_k = ["m = pwAddK(m, '%s', %e);" % (p.name, p.value)
                for p in self.model.parameters]

        # equations (one for each dynamic variable)
        # Note that we just generate C code, which for basic math expressions
        # is identical to matlab.  We just have to change 'pow' to 'power'.
        # Ideally there would be a matlab formatter for sympy.
        pw_ode = ["m = pwAddODE(m, 's%d', '%s');" %
                  (i, sympy.ccode(self.model.odes[i]))
                  for i in range(len(self.model.odes))]
        pw_ode = [re.sub(r'pow(?=\()', 'power', s) for s in pw_ode]

        # observables
        pw_y = ["m = pwAddY(m, '%s', '%s');" %
                (obs.name,
                    ' + '.join('%f * s%s' % t
                               for t in zip(obs.coefficients, obs.species)))
                 for obs in self.model.observables]

        # Add docstring, if present
        if self.docstring:
            output.write('% ' + self.docstring.replace('\n', '\n% '))
            output.write('\n')

        output.write('% PottersWheel model definition file\n')
        output.write('%% save as %s.m\n' % model_name)
        output.write('function m = %s()\n' % model_name)
        output.write('\n')
        output.write('m = pwGetEmptyModel();\n')
        output.write('\n')
        output.write('% meta information\n')
        output.write("m.ID          = '%s';\n" % model_name)
        output.write("m.name        = '%s';\n" % model_name)
        output.write("m.description = '';\n")
        output.write("m.authors     = {''};\n")
        output.write("m.dates       = {''};\n")
        output.write("m.type        = 'PW-1-5';\n")
        output.write('\n')
        output.write('% dynamic variables\n')
        for x in pw_x:
            output.write(x)
            output.write('\n')
        output.write('\n')
        output.write('% dynamic parameters\n')
        for k in pw_k:
            output.write(k)
            output.write('\n')
        output.write('\n')
        output.write('% ODEs\n')
        for ode in pw_ode:
            output.write(ode)
            output.write('\n')
        output.write('\n')
        output.write('% observables\n')
        for y in pw_y:
            output.write(y)
            output.write('\n')
        output.write('\n')
        output.write('%% end of PottersWheel model %s\n' % model_name)
        return output.getvalue()




"""A version of BAX pore assembly where subunits are added to the growing
complex one at a time (contrast with bax_pore.py). Also implements cargo
transport (Smac).
"""

from __future__ import print_function
from pysb import *
from pysb.macros import assemble_pore_sequential, pore_transport, pore_species

Model()

# s1,s2 are ring-formation sites; t is the transport (cargo binding) site
Monomer('Bax', ['s1', 's2', 't'])
Annotation(Bax, 'http://identifiers.org/uniprot/Q07812')
# loc is the location, (m)itochondrion or (c)ytosol; t is the transport site
Monomer('Smac', ['loc', 't'], {'loc': ['m','c']})
Annotation(Smac, 'http://identifiers.org/uniprot/Q9NR28')

Parameter('Bax_0', 8e4)
Parameter('Smac_0', 1e5)
for p in Bax_0, Smac_0:
    Annotation(p, 'http://identifiers.org/doi/10.1371/journal.pcbi.1002482',
               'isDescribedBy')

# Start with monomeric Bax and Smac in the mitochondrion
Initial(Bax(s1=None, s2=None, t=None), Bax_0)
Initial(Smac(loc='m', t=None), Smac_0)

# Maximum number of subunits in a pore
max_size = 6
# Size at which pores are "competent" to transport cargo
min_transport_size = 4

# Build handy rate "sets" (in this formulation, rates don't vary with pore size)
assembly_rates = [[2e-4, 1e-3]] * (max_size - 1)
transport_rates = [[3e-5, 1e-3, 1e1]] * (max_size - min_transport_size + 1)

# Assemble the pore
# (specify t=None so the pore can't fall apart while it's bound to cargo)
assemble_pore_sequential(Bax(t=None), 's1', 's2', max_size, assembly_rates)
# Transport Smac
pore_transport(Bax, 's1', 's2', 't', min_transport_size, max_size,
               Smac(loc='m'), 't', Smac(loc='c'), transport_rates)

# Add an observable for each pore size
for size in range(1, max_size + 1):
    Observable('Bax%d' % size, pore_species(Bax, 's1', 's2', size))
# Observe unbound Smac in each compartment
Observable('mSmac', Smac(loc='m', t=None))
Observable('cSmac', Smac(loc='c'))


if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid. Please see run_bax_pore_sequential.py for example
usage.""")



import matplotlib.pyplot as plt
import numpy as np
from pysb.simulator.bng import BngSimulator
from kinase_cascade import model

# We will integrate from t=0 to t=40
t = np.linspace(0, 40, 50)

# Simulate the model
print("Simulating...")
sim = BngSimulator(model)
x = sim.run(tspan=t, verbose=False, n_runs=5, method='ssa')
tout = x.tout
y = np.array(x.observables)
plt.plot(tout.T, y['ppMEK'].T)
plt.show()



"""Reproduce figures 4A and 4B from the EARM 1.0 publication (Albeck et
al. 2008)."""

from __future__ import print_function
from pysb.simulator import ScipyOdeSimulator
import matplotlib.pyplot as plt
from numpy import *

from earm_1_0 import model


# saturating level of ligand (corresponding to ~1000 ng/ml SuperKiller TRAIL)
Lsat = 6E4; 

# relationship of ligand concentration in the model (in # molecules/cell) to actual TRAIL concentration (in ng/ml)
Lfactor = model.parameters['L_0'].value / 50;

L_0_baseline = model.parameters['L_0'].value


def fig_4a():
    print("Simulating model for figure 4A...")
    t = linspace(0, 20*3600, 20*60+1)  # 20 hours, in seconds, 1 min sampling
    dt = t[1] - t[0]

    Ls_exp = Lsat / array([1, 4, 20, 100, 500])
    Td_exp = [144.2, 178.7, 236,   362.5, 656.5]
    Td_std = [32.5,   32.2,  36.4,  78.6, 171.6]
    Ts_exp = [21.6,   23.8,  27.2,  22.0,  19.0]
    Ts_std = [9.5,     9.5,  12.9,   7.7,  10.5]

    CVenv = 0.2
    # num steps was originally 40, but 15 is plenty smooth enough for screen display
    Ls = floor(logspace(1,5,15)) 

    fs = empty_like(Ls)
    Ts = empty_like(Ls)
    Td = empty_like(Ls)
    print("Scanning over %d values of L_0" % len(Ls))
    for i in range(len(Ls)):
        model.parameters['L_0'].value = Ls[i]

        print("  L_0 = %g" % Ls[i])
        x = ScipyOdeSimulator(model).run(tspan=t).all

        fs[i] = (x['PARP_unbound'][0] - x['PARP_unbound'][-1]) / x['PARP_unbound'][0]
        dP = 60 * (x['PARP_unbound'][:-1] - x['PARP_unbound'][1:]) / (dt * x['PARP_unbound'][0])  # in minutes
        ttn = argmax(dP)
        dPmax = dP[ttn]
        Ts[i] = 1 / dPmax  # minutes
        Td[i] = t[ttn] / 60  # minutes

    plt.figure("Figure 4A")
    plt.plot(Ls/Lfactor, Td, 'g-', Ls/Lfactor, (1-CVenv)*Td, 'g--', Ls/Lfactor,
          (1+CVenv)*Td, 'g--')
    plt.errorbar(Ls_exp/Lfactor, Td_exp, Td_std, None, 'go', capsize=0),
    plt.ylabel('Td (min)'),
    plt.xlabel('TRAIL (ng/ml)'),
    a = plt.gca()
    a.set_xscale('log')
    a.set_xlim((min(Ls) / Lfactor, max(Ls) / Lfactor))
    a.set_ylim((0, 1000))

    model.parameters['L_0'].value = L_0_baseline


def fig_4b():
    print("Simulating model for figure 4B...")

    t = linspace(0, 6*3600, 6*60+1)  # 6 hours
    x = ScipyOdeSimulator(model).run(tspan=t).all

    x_norm = c_[x['Bid_unbound'], x['PARP_unbound'], x['mSmac_unbound']]
    x_norm = 1 - x_norm / x_norm[0, :]  # gets away without max() since first values are largest

    # this is what I originally thought 4B was plotting. it's actually very close. -JLM
    #x_norm = array([x['tBid_total'], x['CPARP_total'], x['cSmac_total']]).T
    #x_norm /= x_norm.max(0)

    tp = t / 3600  # x axis as hours

    plt.figure("Figure 4B")
    plt.plot(tp, x_norm[:,0], 'b', label='IC substrate (tBid)')
    plt.plot(tp, x_norm[:,1], 'y', label='EC substrate (cPARP)')
    plt.plot(tp, x_norm[:,2], 'r', label='MOMP (cytosolic Smac)')
    plt.legend(loc='upper left', bbox_to_anchor=(0,1)).draw_frame(False)
    plt.xlabel('Time (hr)')
    plt.ylabel('fraction')
    a = plt.gca()
    a.set_ylim((-.05, 1.05))


fig_4a()
fig_4b()
plt.show()



import os
import pysb.pathfinder as pf
from pysb.bng import BngFileInterface
from pysb.importers.bngl import model_from_bngl, BnglImportError
from pysb.importers.sbml import model_from_sbml, model_from_biomodels
import numpy
from nose.tools import assert_raises_regexp, raises
import warnings
import mock
import tempfile
import shutil
from pysb.logging import get_logger

# Some models don't match BNG originals exactly due to loss of numerical
# precision. See https://github.com/pysb/pysb/issues/443
REDUCED_PRECISION = {
    'CaOscillate_Func': 1e-4,
    'michment': 1e-8,
    'motor': 1e-8,
    'fceri_ji': 1e-4,
    'test_paramname': 1e-4,
    'tlmr': 1e-4
}

logger = get_logger(__name__)


def bngl_import_compare_simulations(bng_file, force=False,
                                    precision=1e-12,
                                    sim_times=range(0, 100, 10)):
    """
    Test BNGL file import by running an ODE simulation on the imported model
    and on the BNGL file directly to compare trajectories.
    """
    m = model_from_bngl(bng_file, force=force)

    if sim_times is None:
        # Skip simulation check
        return

    # Simulate using the BNGL file directly
    with BngFileInterface(model=None) as bng:
        bng.action('readFile', file=bng_file, skip_actions=1)
        bng.action('generate_network')
        bng.action('simulate', method='ode', sample_times=sim_times)
        bng.execute()
        yfull1 = bng.read_simulation_results()

    # Convert to a PySB model, then simulate using BNG
    with BngFileInterface(model=m) as bng:
        bng.action('generate_network')
        bng.action('simulate', method='ode', sample_times=sim_times)
        bng.execute()
        yfull2 = bng.read_simulation_results()

    # Don't check trajectories on forced examples
    if force:
        return

    assert len(yfull1.dtype.names) == len(yfull2.dtype.names)
    for species in yfull1.dtype.names:
        logger.debug(species)
        logger.debug(yfull1[species])
        if species in yfull2.dtype.names:
            renamed_species = species
        else:
            renamed_species = 'Obs_{}'.format(species)
        logger.debug(yfull2[renamed_species])
        assert numpy.allclose(yfull1[species], yfull2[renamed_species],
                              atol=precision, rtol=precision)


def bngl_import_compare_nfsim(bng_file):
    m = model_from_bngl(bng_file)

    BNG_SEED = 123

    # Simulate using the BNGL file directly
    with BngFileInterface(model=None) as bng:
        bng.action('readFile', file=bng_file, skip_actions=1)
        bng.action('simulate', method='nf', n_steps=10, t_end=100,
                   seed=BNG_SEED)
        bng.execute()
        yfull1 = bng.read_simulation_results()

    # Convert to a PySB model, then simulate using BNG
    with BngFileInterface(model=m) as bng:
        bng.action('simulate', method='nf', n_steps=10, t_end=100,
                   seed=BNG_SEED)
        bng.execute()
        yfull2 = bng.read_simulation_results()

    # Check all species trajectories are equal (within numerical tolerance)
    for i in range(len(m.observables)):
        print(i)
        print(yfull1[i])
        print(yfull2[i])
        print(yfull1[i] == yfull2[i])
        assert yfull1[i] == yfull2[i]


def _bng_validate_directory():
    """ Location of BNG's validation models directory"""
    bng_exec = os.path.realpath(pf.get_path('bng'))
    if bng_exec.endswith('.bat'):
        conda_prefix = os.environ.get('CONDA_PREFIX')
        if conda_prefix:
            return os.path.join(conda_prefix, 'share\\bionetgen\\Validate')

    return os.path.join(os.path.dirname(bng_exec), 'Validate')


def _bngl_location(filename):
    """
    Gets the location of one of BioNetGen's validation model files in BNG's
    Validate directory.
    """
    bngl_file = os.path.join(_bng_validate_directory(), filename + '.bngl')
    return bngl_file


def _sbml_location(filename):
    """
    Gets the location of one of BioNetGen's validation SBML files in BNG's
    Validate/INPUT_FILES directory.
    """
    sbml_file = os.path.join(
        _bng_validate_directory(), 'INPUT_FILES', filename + '.xml')
    return sbml_file


def test_bngl_import_expected_passes_with_force():
    for filename in ('Haugh2b',
                     'Motivating_example',
                     ):
        full_filename = _bngl_location(filename)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            yield (bngl_import_compare_simulations, full_filename, True)


def test_bngl_import_expected_passes_nfsim():
    for filename in ('isingspin_localfcn', ):
        full_filename = _bngl_location(filename)
        yield (bngl_import_compare_nfsim, full_filename)


def test_bngl_import_expected_passes_no_sim():
    """ These models convert properly, but we cannot generate network """
    for filename in ('blbr',         # Uses max_stoich option for netgen
                     'hybrid_test',  # Population maps are not converted
                     'tlbr'):        # Uses max_iter option for netgen
        full_filename = _bngl_location(filename)
        yield (bngl_import_compare_simulations, full_filename, False, None,
               None)


def test_bngl_import_expected_passes():
    for filename in ('CaOscillate_Func',
                     'continue',
                     'deleteMolecules',
                     'egfr_net',
                     'empty_compartments_block',
                     'gene_expr',
                     'gene_expr_func',
                     'gene_expr_simple',
                     'isomerization',
                     'localfunc',
                     'michment',
                     'Motivating_example_cBNGL',
                     'motor',
                     'simple_system',
                     'test_compartment_XML',
                     'test_setconc',
                     'test_synthesis_cBNGL_simple',
                     'test_synthesis_complex',
                     'test_synthesis_complex_0_cBNGL',
                     'test_synthesis_complex_source_cBNGL',
                     'test_synthesis_simple',
                     'toy-jim',
                     'univ_synth',
                     'visualize',
                     'Repressilator',
                     'fceri_ji',
                     'test_paramname',
                     'tlmr'):
        full_filename = _bngl_location(filename)
        yield (bngl_import_compare_simulations, full_filename, False,
               REDUCED_PRECISION.get(filename, 1e-12))


def test_bngl_import_expected_errors():
    errtype = {'ratelawtype': 'Rate law \w* has unknown type',
               'ratelawmissing': 'Rate law missing for rule',
               'plusminus': 'PLUS/MINUS state values',
               'statelabels': 'BioNetGen component/state labels are not yet supported',
              }
    expected_errors = {'ANx': errtype['plusminus'],
                       'CaOscillate_Sat': errtype['ratelawtype'],
                       'heise': errtype['statelabels'],
                       'isingspin_energy': errtype['ratelawmissing'],
                       'test_MM': errtype['ratelawtype'],
                       'test_sat': errtype['ratelawtype'],
                       }

    for filename, errmsg in expected_errors.items():
        full_filename = _bngl_location(filename)
        yield (assert_raises_regexp,
               BnglImportError,
               errmsg,
               bngl_import_compare_simulations,
               full_filename)


def test_sbml_import_flat_model():
    model_from_sbml(_sbml_location('test_sbml_flat_SBML'))


def test_sbml_import_structured_model():
    model_from_sbml(_sbml_location('test_sbml_structured_SBML'), atomize=True)


def _sbml_for_mocks(accession_no, mirror):
    # Need to make a copy because import_from_biomodels deletes the SBML
    # after import
    _, filename = tempfile.mkstemp()
    shutil.copy(_sbml_location('test_sbml_flat_SBML'), filename)
    return filename


@mock.patch('pysb.importers.sbml._download_biomodels', _sbml_for_mocks)
def test_biomodels_import_with_mock():
    model_from_biomodels('1')


@raises(ValueError)
def test_biomodels_invalid_mirror():
    model_from_biomodels('1', mirror='spam')
